from typing import Dict, Any, Callable, List
from .error import FunctionNotExistError
import logging


class LanguageObjectInterface:
    """
    LineUp Object with functions who can be executed by the language
    """
    functions: Dict[str, Callable[..., Any]]
    logger = logging.getLogger("lineup_lang")

    def get_all_functions(self) -> List[str]:
        """
        Get all functions name in the object
        """
        return list(self.functions.keys())

    def execute(self, function_name: str, *args) -> Any:
        """
        Execute a function in the object
        """
        if function_name not in self.functions:
            msg = f"'{function_name}' not exist in '{self}'"
            self.logger.error(msg)
            raise FunctionNotExistError(msg)
        return self.functions[function_name](*args)

    def close(self):
        """
        Destroy the object
        """
        pass

    def reset(self):
        """
        Reset the object for future new exectution
        """
        pass

    def __str__(self) -> str:
        return f"<LUPO:{self.__class__.__name__}>"


class CoreObjectInterface(LanguageObjectInterface):
    """
    Object instancied before the creation of the language executor

    Function in this object is directly accessible by the language executor
    """
    executor: Any
    version: str = "0.0.0"

    def set_executor(self, executor: Any) -> None:
        self.executor = executor

    def get_version(self) -> str:
        return self.version

    def __str__(self) -> str:
        return f"<LUPC:{self.__class__.__name__}>"


class LanguageExecutorInterface:
    """
    Language executor interface

    This object is the main object who execute the script
    It execute line each by each
    """
    _core_function: Dict[str, LanguageObjectInterface]
    _core: List[LanguageObjectInterface]
    _version: str = "0.0.0"

    def execute_line(self, line: List[str]):
        pass

    def execute(self, script: List[List[str]]) -> Any:
        pass

    def get_versions(self) -> Dict[str, str]:
        versions = {self.__str__(): self._version}
        for core in self._core:
            versions[core.__str__()] = core.get_version()
        return versions

    def close(self) -> None:
        for core in self._core:
            logging.getLogger("lineup_lang").info(f"Close: {core}")
            core.close()

    def reset(self) -> None:
        for core in self._core:
            logging.getLogger("lineup_lang").info(f"Reset: {core}")
            core.reset()

    def __str__(self) -> str:
        return f"<LUPE:{self.__class__.__name__}>"


class LanguageInterface:
    """"
    Language interface
    It read the script, cut it in line and
    execute it with the language executor
    """
    def close(self):
        pass

    def get_all_functions(self) -> List[str]:
        pass

    def get_versions(self) -> str:
        pass

    def execute_script(self, script: str) -> Any:
        pass

    def execute_script_with_args(self, script: str, **kwargs) -> Any:
        pass

    def execute_file(self, file_path: str, **kwargs) -> Any:
        pass

    def __str__(self) -> str:
        return f"<LUPL:{self.__class__.__name__}>"



from typing import List, Any, Dict
from .language_object import LanguageInterface, LanguageExecutorInterface, \
     LanguageObjectInterface
from .error import ArgumentNotExistError, DecodeLineStringError, LineupError
from .logger import start_logging
import lineup_lang.executor as luexec
import lineup_lang.core as lucore
import regex as re
import logging
import os

__all__ = ["Language", "LanguageObjectInterface", "luexec", "lucore"]


class Language(LanguageInterface):
    _executor: LanguageExecutorInterface
    no_error: bool
    logger = logging.getLogger("lineup_lang")

    def __init__(self, executor: LanguageExecutorInterface,
                 no_error: bool = True, log_level: str = "WARN"):
        start_logging(log_level)
        self._executor = executor
        self.no_error = no_error

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def _resolve_line(self, line: str):
        lines = line.split(" ")
        result = []
        tmp = ""
        for data in lines:
            if data.startswith('"') and data.endswith('"'):
                result.append("".join(data[1:-1]))
            elif data.startswith('"'):
                if tmp:
                    raise DecodeLineStringError(
                        f"'{line}' is not valid line string")
                tmp = data
            elif data.endswith('"'):
                if not tmp:
                    raise DecodeLineStringError(
                        f"'{line}' is not valid line string")
                tmp += " " + data
                result.append("".join(tmp[1:-1]))
                tmp = ""
            elif tmp:
                tmp += " " + data
            else:
                result.append(data)
        if tmp:
            raise DecodeLineStringError(
                f"'{line}' is not valid line string")
        return result

    def _get_line(self, line: str) -> List[str] | None:
        line = line.strip()
        if not line:
            return None
        if line.startswith("#"):
            return None
        return self._resolve_line(line)

    def _resolve_args(self, script: str, **kwargs):
        regex = r"\$(\((\w+):(.+?)\)|(\w+))"
        matches = re.finditer(regex, script)
        for match in matches:
            keyname = match.group(2) or match.group(4)
            default_value = match.group(3)
            if keyname in kwargs:
                value = kwargs[keyname]
            elif default_value is None:
                raise ArgumentNotExistError(
                    f"'{keyname}' not exist in '{kwargs}'")
            else:
                value = default_value
            script = script.replace(match.group(0), "\"" + value + "\"")
        return script

    def close(self):
        self._executor.close()

    def get_all_functions(self) -> List[str]:
        return self._executor.get_all_functions()
    
    def get_versions(self) -> Dict[str, str]:
        file_version = os.path.join(os.path.dirname(__file__), "VERSION")
        lup_version = "0.0.0"
        with open(file_version, "r") as file:
            lup_version = file.read().strip()
        all_versions = self._executor.get_versions()
        all_versions["lineup_lang"] = lup_version
        for [name, version] in all_versions.items():
            if version is None:
                all_versions[name] = lup_version
        return all_versions

    def execute_script(self, script: str) -> Any:
        self.logger.debug(f"Execute script:\n{script}")
        script_lines = []
        for line in script.split("\n"):
            line = self._get_line(line)
            if line:
                script_lines.append(line)
        try:
            result = self._executor.execute(script_lines)
        except LineupError as e:
            if self.no_error:
                return e
            raise e
        self._executor.reset()
        return result

    def execute_script_with_args(self, script: str, **kwargs) -> Any:
        script = self._resolve_args(script, **kwargs)
        return self.execute_script(script)

    def execute_file(self, file_path: str, **kwargs) -> Any:
        with open(file_path, "r") as file:
            script = file.read()
        return self.execute_script_with_args(script, **kwargs)

    def __str__(self) -> str:
        return f"<{self.__class__.__name__}>"



class LineupError(Exception):
    pass


class ExecutorFunctionAlreadyExistError(LineupError):
    pass


class FunctionNotExistError(LineupError):
    pass


class ExecutorFunctionNotExistError(FunctionNotExistError):
    pass


class ArgumentNotExistError(LineupError):
    pass


class DecodeLineStringError(LineupError):
    pass
